// Ripped from https://github.com/Ch0pin/medusa/ and modified to fit Androguard packets

colorLog('[+] LOADING FRAMEWORK/CORDOVA.JS',{c: Color.Red});

var pluginManager = tryGetClass("org.apache.cordova.PluginManager");

if (pluginManager) {
    pluginManager.instantiatePlugin.implementation = function(pluginName){
        agPacket({pluginName: pluginName}).send();
        return this.instantiatePlugin(pluginName);
    }    
}

